import math

##WGS84 Point
class WGS84Point() :
	##Constructor
	##@param latitude : wgs84 latitude
	##@param longitude : wgs84 longitude
	##@param eph : gps horizontal precision
	##@param eps : gps speed precision
	##@param epd : gps direction precision
	def __init__(self, latitude=0, longitude=0, eph=0.0, eps=0.0, epd=0.0):
		self.latitude = latitude
		self.longitude = longitude
		self.eph = eph
		self.eps = eps
		self.epd = epd
		
	##Copy object content
	def copy(self, other):
		self.latitude = other.latitude
		self.longitude = other.longitude
		self.eph = other.eph
		self.eps = other.eps
		self.epd = other.epd
	
	##To string
	def __str__(self):
		return 'WGS84Point(lat=%.5f long=%.5f, eph=%.3f, eps=%.3f, epd=%.3f)' % (self.latitude, self.longitude, self.eph, self.eps, self.epd)

##UTM Point (projection)
class UTMPoint() :
	##Constructor
	##@param x : x
	##@param y : y
	##@param zone : UTM zone
	##@northHemisphere : True if point in north hemisphere
	def __init__(self, x=0, y=0, zone=0, northHemisphere=True):
		self.x = x
		self.y = y
		self.zone = zone
		self.northHemisphere = northHemisphere
		
	##Copy object content
	def copy(self, other):
		self.x = other.x
		self.y = other.y
		self.zone = other.zone
		self.northHemisphere = other.northHemisphere
	
	##To string
	def __str__(self):
		hemisphere = 'S'
		if self.northHemisphere:
			hemisphere = 'N'
		return 'UTMPoint(x=%.5f y=%.5f zone=%d hemisphere=%s)' % (self.x, self.y, self.zone, hemisphere)

##UTM rectangle contains 4 UTM points to define a rectangle
class UTMRectangle():
	##Constructor
	##@param topLeft : top left point
	##@param topRight : top right point
	##@param bottomLeft : bottom left point
	##@param bottomRight : bottom right point
	def __init__(self, topLeft=None, topRight=None, bottomLeft=None, bottomRight=None):
		#members
		self.topLeft = None
		self.topRight = None
		self.bottomLeft = None
		self.bottomRight = None
	
		if topLeft==None:
			self.topLeft = UTMPoint()
		else:
			self.topLeft = topLeft
		if topRight==None:
			self.topRight = UTMPoint()
		else:
			self.topRight = topRight
		if bottomLeft==None:
			self.bottomLeft = UTMPoint()
		else:
			self.bottomLeft = bottomLeft
		if bottomRight==None:
			self.bottomRight = UTMPoint()
		else:
			self.bottomRight = bottomRight
	
	##Check if point is in rectangle
	def isInRectangle(self, utm):
		if self.topLeft.x<=utm.x and utm.x<=self.topRight.x and self.bottomLeft.y<=utm.y and utm.y<=self.topLeft.y :
			return True
		else:
			return False
		
	##To string
	def __str__(self):
		return 'UTMRectangle TL(x=%.5f y=%.5f) TR(x=%.5f y=%.5f) BL(x=%.5f y=%.5f) BR(x=%.5f y=%.5f)' % (self.topLeft.x, self.topLeft.y, self.topRight.x, self.topRight.y, self.bottomLeft.x, self.bottomLeft.y, self.bottomRight.x, self.bottomRight.y)

##Converts coordinates from UTM to Geographic WGS84 and from WGS84 to UTM
##Sources from javascript http://home.hiwaay.net/~taylorc/toolbox/geography/geoutm.html
##@info Geographic coordinates are entered and displayed in decimal degrees. Negative 
##      numbers indicate West longitudes and South latitudes. UTM coordinates are entered and displayed in meters. 
##      The ellipsoid model used for computations is WGS84.
class GISConvert() :
	##Ellipsoid model constants (actual values here are for WGS84)
	sm_a = 6378137.0
	sm_b = 6356752.314
	UTMScaleFactor = 0.9996
	
	##Converts degrees to radians.
	def degToRad(deg):
		return (deg / 180.0 * math.pi)
	degToRad = staticmethod(degToRad)

	##Converts radians to degrees.
	def radToDeg(rad):
		return (rad / math.pi * 180.0)
	radToDeg = staticmethod(radToDeg)

	##Computes the ellipsoidal distance from the equator to a point at a
	##given latitude.
	##Reference: Hoffmann-Wellenhof, B., Lichtenegger, H., and Collins, J.,
	##GPS: Theory and Practice, 3rd ed.  New York: Springer-Verlag Wien, 1994.
	##Inputs:
	##    phi - Latitude of the point, in radians.
	##Globals:
	##    self.GISConvert.sm_a - Ellipsoid model major axis.
	##    self.GISConvert.sm_b - Ellipsoid model minor axis.
	##Returns:
	##    The ellipsoidal distance of the point from the equator, in meters.
	def arcLengthOfMeridian(phi):
		#Precalculate n
		n = (GISConvert.sm_a - GISConvert.sm_b) / (GISConvert.sm_a + GISConvert.sm_b)
		
		#Precalculate alpha
		alpha = ((GISConvert.sm_a + GISConvert.sm_b) / 2.0) * (1.0 + (math.pow (n, 2.0) / 4.0) + (math.pow (n, 4.0) / 64.0))
		
		#Precalculate beta
		beta = (-3.0 * n / 2.0) + (9.0 * math.pow (n, 3.0) / 16.0) + (-3.0 * math.pow (n, 5.0) / 32.0)
		
		#Precalculate gamma
		gamma = (15.0 * math.pow (n, 2.0) / 16.0) + (-15.0 * math.pow (n, 4.0) / 32.0)
		
		#Precalculate delta
		delta = (-35.0 * math.pow (n, 3.0) / 48.0) + (105.0 * math.pow (n, 5.0) / 256.0)
		
		#Precalculate epsilon
		epsilon = (315.0 * math.pow (n, 4.0) / 512.0)
		
		#Now calculate the sum of the series and return
		result = alpha * (phi + (beta * math.sin (2.0 * phi)) + (gamma * math.sin (4.0 * phi)) + (delta * math.sin (6.0 * phi)) + (epsilon * math.sin (8.0 * phi)))
		
		return result
	arcLengthOfMeridian = staticmethod(arcLengthOfMeridian)
	
	##Determines the central meridian for the given UTM zone.
	##Inputs:
	##    zone - An integer value designating the UTM zone, range [1,60].
	##Returns:
	##  The central meridian for the given UTM zone, in radians, or zero
	##  if the UTM zone parameter is outside the range [1,60].
	##  Range of the central meridian is the radian equivalent of [-177,+177].
	def utmCentralMeridian(zone):
		cmeridian = GISConvert.degToRad(-183.0 + (zone * 6.0))
		
		return cmeridian
	utmCentralMeridian = staticmethod(utmCentralMeridian)
	
	##Computes the footpoint latitude for use in converting transverse
	##Mercator coordinates to ellipsoidal coordinates.
	##Reference: Hoffmann-Wellenhof, B., Lichtenegger, H., and Collins, J.,
	##  GPS: Theory and Practice, 3rd ed.  New York: Springer-Verlag Wien, 1994.
	##Inputs:
	##  y - The UTM northing coordinate, in meters.
	##Returns:
	##  The footpoint latitude, in radians.
	def footpointLatitude(y):
		#Precalculate n (Eq. 10.18)
		n = (GISConvert.sm_a - GISConvert.sm_b) / (GISConvert.sm_a + GISConvert.sm_b)
		
		#Precalculate alpha_ (Eq. 10.22)
		#(Same as alpha in Eq. 10.17)
		alpha_ = ((GISConvert.sm_a + GISConvert.sm_b) / 2.0) * (1 + (math.pow (n, 2.0) / 4) + (math.pow (n, 4.0) / 64))
		
		#Precalculate y_ (Eq. 10.23)
		y_ = y / alpha_
		
		#Precalculate beta_ (Eq. 10.22)
		beta_ = (3.0 * n / 2.0) + (-27.0 * math.pow (n, 3.0) / 32.0) + (269.0 * math.pow (n, 5.0) / 512.0)
		
		#Precalculate gamma_ (Eq. 10.22)
		gamma_ = (21.0 * math.pow (n, 2.0) / 16.0) + (-55.0 * math.pow (n, 4.0) / 32.0)
		
		#Precalculate delta_ (Eq. 10.22)
		delta_ = (151.0 * math.pow (n, 3.0) / 96.0) + (-417.0 * math.pow (n, 5.0) / 128.0)
		
		#Precalculate epsilon_ (Eq. 10.22)
		epsilon_ = (1097.0 * math.pow (n, 4.0) / 512.0)
		
		#Now calculate the sum of the series (Eq. 10.21)
		result = y_ + (beta_ * math.sin (2.0 * y_)) + (gamma_ * math.sin (4.0 * y_)) + (delta_ * math.sin (6.0 * y_)) + (epsilon_ * math.sin (8.0 * y_))
		
		return result
	footpointLatitude = staticmethod(footpointLatitude)
	
	##Converts a latitude/longitude pair to x and y coordinates in the
	##Transverse Mercator projection.  Note that Transverse Mercator is not
	##the same as UTM; a scale factor is required to convert between them.
	##Reference: Hoffmann-Wellenhof, B., Lichtenegger, H., and Collins, J.,
	##GPS: Theory and Practice, 3rd ed.  New York: Springer-Verlag Wien, 1994.
	##Inputs:
	##	phi - Latitude of the point, in radians.
	##   lambda - Longitude of the point, in radians.
	##   lambda0 - Longitude of the central meridian to be used, in radians.
	##Outputs:
	##   xy - A 2-element array containing the x and y coordinates
	##        of the computed point.
	##Returns:
	##   The function does not return a value.
	def mapLatLonToXY(wgs84):
		#init
		phi = GISConvert.degToRad(wgs84.latitude)
		lambda_ = GISConvert.degToRad(wgs84.longitude)
		zone = GISConvert.computeUtmZone(wgs84.longitude)
		lambda0 = GISConvert.utmCentralMeridian(zone)
	
		#Precalculate ep2
		ep2 = (math.pow(GISConvert.sm_a, 2.0) - math.pow(GISConvert.sm_b, 2.0)) / math.pow(GISConvert.sm_b, 2.0)
		
		#Precalculate nu2
		nu2 = ep2 * math.pow(math.cos (phi), 2.0)
		
		#Precalculate N
		N = math.pow(GISConvert.sm_a, 2.0) / (GISConvert.sm_b * math.sqrt(1 + nu2))
		
		#Precalculate t
		t = math.tan(phi)
		t2 = t * t
		#double tmp = (t2 * t2 * t2) - math.pow (t, 6.0);
		
		#Precalculate l
		l = lambda_ - lambda0
		
		#Precalculate coefficients for l**n in the equations below
		#so a normal human being can read the expressions for easting
		#and northing
		# -- l**1 and l**2 have coefficients of 1.0 
		l3coef = 1.0 - t2 + nu2
		l4coef = 5.0 - t2 + 9 * nu2 + 4.0 * (nu2 * nu2)
		l5coef = 5.0 - 18.0 * t2 + (t2 * t2) + 14.0 * nu2 - 58.0 * t2 * nu2
		l6coef = 61.0 - 58.0 * t2 + (t2 * t2) + 270.0 * nu2 - 330.0 * t2 * nu2
		l7coef = 61.0 - 479.0 * t2 + 179.0 * (t2 * t2) - (t2 * t2 * t2)
		l8coef = 1385.0 - 3111.0 * t2 + 543.0 * (t2 * t2) - (t2 * t2 * t2)
		
		#Calculate easting (x)
		point = UTMPoint()
		point.zone = zone
		if wgs84.latitude<0:
			point.northHemisphere = False
		else:
			point.northHemisphere = True
		point.x = N * math.cos (phi) * l \
				+ (N / 6.0 * math.pow(math.cos (phi), 3.0) * l3coef * math.pow(l, 3.0)) \
				+ (N / 120.0 * math.pow(math.cos (phi), 5.0) * l5coef * math.pow(l, 5.0)) \
				+ (N / 5040.0 * math.pow(math.cos (phi), 7.0) * l7coef * math.pow(l, 7.0))
		
		#Calculate northing (y)
		point.y = GISConvert.arcLengthOfMeridian(phi) \
				+ (t / 2.0 * N * math.pow(math.cos (phi), 2.0) * math.pow(l, 2.0)) \
				+ (t / 24.0 * N * math.pow(math.cos (phi), 4.0) * l4coef * math.pow(l, 4.0)) \
				+ (t / 720.0 * N * math.pow(math.cos (phi), 6.0) * l6coef * math.pow(l, 6.0)) \
				+ (t / 40320.0 * N * math.pow(math.cos (phi), 8.0) * l8coef * math.pow(l, 8.0))
			
		return point
	mapLatLonToXY = staticmethod(mapLatLonToXY)
		
	##Converts x and y coordinates in the Transverse Mercator projection to
	##a latitude/longitude pair.  Note that Transverse Mercator is not
	##the same as UTM; a scale factor is required to convert between them.
	##
	##Reference: Hoffmann-Wellenhof, B., Lichtenegger, H., and Collins, J.,
	##  GPS: Theory and Practice, 3rd ed.  New York: Springer-Verlag Wien, 1994.
	##
	##Inputs:
	##  x - The easting of the point, in meters.
	##  y - The northing of the point, in meters.
	##  lambda0 - Longitude of the central meridian to be used, in radians.
	##
	##Outputs:
	##  philambda - A 2-element containing the latitude and longitude
	##              in radians.
	##
	##Returns:
	##  The function does not return a value.
	##
	##Remarks:
	##  The local variables Nf, nuf2, tf, and tf2 serve the same purpose as
	##  N, nu2, t, and t2 in MapLatLonToXY, but they are computed with respect
	##  to the footpoint latitude phif.
	##
	##  x1frac, x2frac, x2poly, x3poly, etc. are to enhance readability and
	##  to optimize computations.
	def mapXYToLatLon(utm):
		#compute meridian
		cmeridian = GISConvert.utmCentralMeridian(utm.zone)
	
		#Get the value of phif, the footpoint latitude.
		phif = GISConvert.footpointLatitude(utm.y)
		
		#Precalculate ep2
		ep2 = (math.pow(GISConvert.GISConvert.sm_a, 2.0) - math.pow(GISConvert.GISConvert.sm_b, 2.0)) / math.pow(GISConvert.GISConvert.sm_b, 2.0)
		
		#Precalculate cos (phif)
		cf = math.cos(phif)
		
		#Precalculate nuf2
		nuf2 = ep2 * math.pow(cf, 2.0)
		
		#Precalculate Nf and initialize Nfpow
		Nf = math.pow(GISConvert.GISConvert.sm_a, 2.0) / (GISConvert.GISConvert.sm_b * math.sqrt(1 + nuf2))
		Nfpow = Nf
		
		#Precalculate tf
		tf = math.tan(phif)
		tf2 = tf * tf
		tf4 = tf2 * tf2
		
		#Precalculate fractional coefficients for x**n in the equations
		#below to simplify the expressions for latitude and longitude.
		x1frac = 1.0 / (Nfpow * cf)
		
		Nfpow *= Nf;   #now equals Nf**2)
		x2frac = tf / (2.0 * Nfpow)
		
		Nfpow *= Nf;   #now equals Nf**3)
		x3frac = 1.0 / (6.0 * Nfpow * cf)
		
		Nfpow *= Nf;   #now equals Nf**4)
		x4frac = tf / (24.0 * Nfpow)
		
		Nfpow *= Nf;   #now equals Nf**5)
		x5frac = 1.0 / (120.0 * Nfpow * cf)
		
		Nfpow *= Nf;   #now equals Nf**6)
		x6frac = tf / (720.0 * Nfpow)
		
		Nfpow *= Nf;   #now equals Nf**7)
		x7frac = 1.0 / (5040.0 * Nfpow * cf)
		
		Nfpow *= Nf;   #now equals Nf**8)
		x8frac = tf / (40320.0 * Nfpow)
		
		#Precalculate polynomial coefficients for x**n.
		# -- x**1 does not have a polynomial coefficient.
		x2poly = -1.0 - nuf2
		
		x3poly = -1.0 - 2 * tf2 - nuf2
		
		x4poly = 5.0 + 3.0 * tf2 + 6.0 * nuf2 - 6.0 * tf2 * nuf2 - 3.0 * (nuf2 *nuf2) - 9.0 * tf2 * (nuf2 * nuf2)
		
		x5poly = 5.0 + 28.0 * tf2 + 24.0 * tf4 + 6.0 * nuf2 + 8.0 * tf2 * nuf2
		
		x6poly = -61.0 - 90.0 * tf2 - 45.0 * tf4 - 107.0 * nuf2 + 162.0 * tf2 * nuf2
		
		x7poly = -61.0 - 662.0 * tf2 - 1320.0 * tf4 - 720.0 * (tf4 * tf2)
		
		x8poly = 1385.0 + 3633.0 * tf2 + 4095.0 * tf4 + 1575 * (tf4 * tf2)
		
		#Calculate latitude
		point = WGS84Point()
		point.latitude = phif + x2frac * x2poly * (utm.x * utm.x) \
								+ x4frac * x4poly * math.pow(utm.x, 4.0) \
								+ x6frac * x6poly * math.pow(utm.x, 6.0) \
								+ x8frac * x8poly * math.pow(utm.x, 8.0)
		
		#Calculate longitude
		point.longitude = cmeridian + x1frac * utm.x \
								+ x3frac * x3poly * math.pow(utm.x, 3.0) \
								+ x5frac * x5poly * math.pow(utm.x, 5.0) \
								+ x7frac * x7poly * math.pow(utm.x, 7.0)
								
		#convert to radian
		point.latitude = GISConvert.radToDeg(point.latitude)
		point.longitude = GISConvert.radToDeg(point.longitude)
		
		return point
	mapXYToLatLon = staticmethod(mapXYToLatLon)
	
	##Converts a latitude/longitude pair to x and y coordinates in the
	##Universal Transverse Mercator projection.
	##
	##Inputs:
	##  lat - Latitude of the point, in radians.
	##  lon - Longitude of the point, in radians.
	##  zone - UTM zone to be used for calculating values for x and y.
	##         If zone is less than 1 or greater than 60, the routine
	##         will determine the appropriate zone from the value of lon.
	##
	##Outputs:
	##  xy - A 2-element array where the UTM x and y values will be stored.
	##
	##Returns:
	##  The UTM zone used for calculating the values of x and y.
	def wgs84ToUtm(wgs84) :
		#convert
		utm = GISConvert.mapLatLonToXY(wgs84)
		
		#Adjust easting and northing for UTM system.
		utm.x = utm.x * GISConvert.UTMScaleFactor + 500000.0
		utm.y = utm.y * GISConvert.UTMScaleFactor
		if utm.y<0.0 :
			utm.y = utm.y + 10000000.0
			
		return utm
	wgs84ToUtm = staticmethod(wgs84ToUtm)
	
	##Converts x and y coordinates in the Universal Transverse Mercator
	##projection to a latitude/longitude pair.
	##
	##Inputs:
	##	x - The easting of the point, in meters.
	##	y - The northing of the point, in meters.
	##	zone - The UTM zone in which the point lies.
	##	southhemi - True if the point is in the southern hemisphere;
	##              false otherwise.
	##
	##Outputs:
	##	latlon - A 2-element array containing the latitude and
	##           longitude of the point, in radians.
	##
	##Returns:
	##	The function does not return a value.
	def utmToWgs84(utm):
		utm.x -= 500000.0
		utm.x /= GISConvert.__UTMScaleFactor
		
		#If in southern hemisphere, adjust y accordingly.
		if not utm.northHemisphere :
			utm.y -= 10000000.0;
		utm.y /= GISConvert.__UTMScaleFactor

		wgs84 = GISConvert.mapXYToLatLon(utm)
		
		return wgs84
	utmToWgs84 = staticmethod(utmToWgs84)
	
	##Compute UTM zone
	##@param lon Longitude in Degrees
	##@return UTM Zone
	def computeUtmZone(lon):
		return math.floor((lon + 180.0) / 6) + 1
	computeUtmZone = staticmethod(computeUtmZone)

##GISComputeException class
class GISComputeException(Exception):
	def __init__(self, message):
		self.message = message
	
	def __str__(self):
		return self.message

##GISCompute class
class GISCompute():
	##Compute distance between 2 UTM points
	def distanceUtm(utm1, utm2):
		#check if in same zone
		if utm1.zone!=utm2.zone:
			raise GISComputeException('Unable to compute distance from UTM point of different zones')
			
		#compute distance
		return math.sqrt( math.pow((utm2.x - utm1.x), 2) + math.pow((utm2.y - utm1.y), 2))
	distanceUtm = staticmethod(distanceUtm)
	
	##Compute distance between 2 WGS84 points
	def distanceWgs84(wgs841, wgs842):
		#init
		a = 6378137
		b = 6356752.3142
		f = 1/298.257223563 #WGS-84 ellipsoid

		L = math.radians(wgs842.longitude-wgs841.longitude)
		U1 = math.atan((1-f) * math.tan(math.radians(wgs841.latitude)))
		U2 = math.atan((1-f) * math.tan(math.radians(wgs842.latitude)))
		sinU1 = math.sin(U1)
		cosU1 = math.cos(U1)
		sinU2 = math.sin(U2)
		cosU2 = math.cos(U2)
		
		lambda_ = L
		lambdaP = 0
		iterLimit = 20
		
		while True:
			sinLambda = math.sin(lambda_)
			cosLambda = math.cos(lambda_)
			sinSigma = math.sqrt((cosU2*sinLambda) * (cosU2*sinLambda) + (cosU1*sinU2-sinU1*cosU2*cosLambda) * (cosU1*sinU2-sinU1*cosU2*cosLambda))
			if sinSigma==0:
				#co-incident points
				return 0 
			cosSigma = sinU1*sinU2 + cosU1*cosU2*cosLambda
			sigma = math.atan2(sinSigma, cosSigma)
			sinAlpha = cosU1 * cosU2 * sinLambda / sinSigma
			cosSqAlpha = 1 - sinAlpha*sinAlpha
			cos2SigmaM = 0
			try:
				cos2SigmaM = cosSigma - 2*sinU1*sinU2/cosSqAlpha
			except ZeroDivisionError:
				#equatorial line: cosSqAlpha=0
				cos2SigmaM = 0
			except:
				cos2SigmaM = 0
			C = f/16*cosSqAlpha*(4+f*(4-3*cosSqAlpha))
			lambdaP = lambda_
			lambda_ = L + (1-C) * f * sinAlpha * (sigma + C*sinSigma*(cos2SigmaM+C*cosSigma*(-1+2*cos2SigmaM*cos2SigmaM)))
			
			#stop statement
			if abs(lambda_-lambdaP) > 1e-12 and --iterLimit>0:
				break
				
		if iterLimit==0:
			#formula failed to converge
			return 0
			
		uSq = cosSqAlpha * (a*a - b*b) / (b*b)
		A = 1 + uSq/16384*(4096+uSq*(-768+uSq*(320-175*uSq)))
		B = uSq/1024 * (256+uSq*(-128+uSq*(74-47*uSq)))
		deltaSigma = B*sinSigma*(cos2SigmaM+B/4*(cosSigma*(-1+2*cos2SigmaM*cos2SigmaM)-B/6*cos2SigmaM*(-3+4*sinSigma*sinSigma)*(-3+4*cos2SigmaM*cos2SigmaM)))
		s = b*A*(sigma-deltaSigma)
  
		#round to 1mm precision
		s = round(s, 3)
		
		return s
	distanceWgs84 = staticmethod(distanceWgs84)
