from xml.dom.minidom import parse
import datetime
import gtk

##PrefFile class internal object
##Used to store section data from xml file
class PrefFileSection():

    ##Constructor
	def __init__(self):
		self.id = ''
		self.keys = {}

    ##Return section xml string
	def xml(self):
		#init
		result = ''

		#start section
		result += '<section id="%s">\n' % (self.id)
		#add all keys
		for key in self.keys:
			result += '\t' + self.keys[key].xml() + '\n'
		#end section
		result += '</section>\n'

		return result

    ##Return section under string format
	def __str__(self):
		return 'PrefFileSection(id=%s, keys=%d)' % (self.id, len(self.keys))

##PrefFile class internal object
##Used to store key data from xml file
class PrefFileKey():
	##Constructor
	def __init__(self, item):
		self.item = item

	##Return key xml string
	def xml(self):
		return '<key id="%s" type="%s">%s</key>' % (self.item.getId(), self.item.getType(), self.item.getValue())

	##To string
	def __str__(self):
		return 'PrefFileKey(id=%s, type=%s, value=%s)' % (self.item.getId(), self.item.getTtype(), self.item.getValue())

##PrefFileException class
##exception launched by PrefFile object
class PrefFileException(Exception):
	def __init__(self, message):
		Exception.__init__(self, message)

##PrefFile class
##Class to manage preferences xml file
class PrefFile():
	##Constructor
	##@param file : preferences xml file to open
	def __init__(self, file):
		#set members
		self.__sections = {}
		self.__file = file

		#read file
		self.__read()
		
	#Return full filepath
	def getPrefFilePath(self):
		return self.__file;

	##Read pref file
	##@except PrefFileException : when problem occured during pref file reading
	def __read(self):
		try:
			#open document
			document = parse(self.__file)

			#get root
			root = document.documentElement

			#walk through all sections
			sections = root.getElementsByTagName("section")
			for section in sections:
				#create section
				currentSection = PrefFileSection()

				#set section infos
				currentSection.id = section.getAttribute('id')

				#walk through section keys
				keys = section.getElementsByTagName('key')
				for key in keys:
					#get key infos
					id = key.getAttribute('id')
					type = key.getAttribute('type')
					value = ''
					if len(key.childNodes)>0:
						value = key.childNodes[0].nodeValue
				
					#create pref file item
					if type==PrefFileItem.TYPE_STRING:
						item = PrefFileItemString(currentSection.id, id, type, value)
					elif type==PrefFileItem.TYPE_INT:
						item = PrefFileItemInt(currentSection.id, id, type, value)
					elif type==PrefFileItem.TYPE_FLOAT:
						item = PrefFileItemFloat(currentSection.id, id, type, value)
					elif type==PrefFileItem.TYPE_DATE:
						item = PrefFileItemDate(currentSection.id, id, type, value)
					elif type==PrefFileItem.TYPE_COLOR:
						item = PrefFileItemColor(currentSection.id, id, type, value)
					elif type==PrefFileItem.TYPE_BOOLEAN:
						item = PrefFileItemBoolean(currentSection.id, id, type, value)
					else:
						#unknown type
						raise PrefFileException('Unknown item type "%s"' % type)
					
					#save key to section
					currentKey = PrefFileKey(item)
					currentSection.keys[id] = currentKey

				#save section
				self.__sections[currentSection.id] = currentSection

		except Exception, e:
			raise PrefFileException('Unable to read pref file : %s' % e)

	##Save pref file
	##@param file : if specified, save preferences to specified file. Otherwise save it in loaded file
	def save(self, file=''):
		#build new xml string
		xml = '<config>\n'
		for section in self.__sections:
			xml += self.__sections[section].xml()
		xml += '</config>'

		#open file for writing
		if len(file)==0:
			file = self.__file
		f = open(file, 'wb')
		f.write(xml)
		f.close()

	##Get pref item
	##@param sectionId : get item of specified section id
	##@param keyId : get item of specified key id
	##@return : PrefFileItem or None if not found
	def getItem(self, sectionId, keyId):
		#init
		result = None

		#search item
		if self.__sections.has_key(sectionId):
			#section exists
			if self.__sections[sectionId].keys.has_key(keyId):
				result = self.__sections[sectionId].keys[keyId].item
			else:
				#key doesn't exist
				result = None
		else:
			#section doesn't exist
			result = None

		return result

	##Set pref item
	##@param item : PrefFileItem instance to save in preferences
	##@result : True if item updated, False otherwise
	##@except : if item is not PrefFileItem raise PrefFileException
	def setItem(self, item):
		#init
		result = False

		if isinstance(item, PrefFileItem):
			#type is correct, update value
			if self.__sections.has_key(item.getSectionId()):
				#section exists
				if self.__sections[item.getSectionId()].keys.has_key(item.getId()):
					#key exists
					key = self.__sections[item.getSectionId()].keys[item.getId()]
					#update value
					key.value = item.getValue()
					result = True
				else:
					#unknown key id
					result = False
			else:
				#unknown section id
				result = False
		else:
			raise PrefFileException('PrefFileItem object required!')

		return result

##Exception raised by PrefFileItem
class PrefFileItemException(Exception):
	def __init__(self, message):
		Exception.__init__(self, message)

##Base class of item in preferences file
class PrefFileItem():
	##Constants
	TYPE_STRING = 'string'
	TYPE_INT = 'int'
	TYPE_FLOAT = 'float'
	TYPE_DATE = 'date'
	TYPE_COLOR = 'color'
	TYPE_BOOLEAN = 'bool'

	##Constructor
	##@param sectionId : item section id
	##@param id : item key id
	##@param type : item key type
	##@param value : item value
	def __init__(self, sectionId, id, type, value):
		#set section id
		self.__sectionId = sectionId
	
		#set id
		self.__id = id
		
		#set type
		if type==PrefFileItem.TYPE_STRING or type==PrefFileItem.TYPE_INT or type==PrefFileItem.TYPE_FLOAT or type==PrefFileItem.TYPE_DATE or type==PrefFileItem.TYPE_COLOR or type==PrefFileItem.TYPE_BOOLEAN:
			self.__type = type
		else:
			raise PrefFileItemException('Unknown item type')
			
		#set value
		self.setValue(value)
		
	##Id setter
	def __setId(self, id):
		self.__id = id

    ##Id getter
	def getId(self):
		return self.__id

	##Type setter
	def __setType(self, type):
		#check if type exists
		if type==PrefFileItem.TYPE_STRING or type==PrefFileItem.TYPE_INT or type==PrefFileItem.TYPE_FLOAT or type==PrefFileItem.TYPE_DATE or type==PrefFileItem.TYPE_COLOR or type==PrefFileItem.TYPE_BOOLEAN:
			self.__type = type
		else:
			raise PrefFileItemException('Unknown item type')

    ##Type getter
	def getType(self):
		return self.__type

	##Value setter
    ##must be overriden
	def setValue(self, value):
		raise PrefFileItemException('setValue() is not implemented in %s class' % self.__class__.__name__)

    ##Value getter
	def getValue(self):
		return self._value

	##Section id setter
	def __setSectionId(self, sectionId):
		self.__sectionId = sectionId

    ##Section id getter
	def getSectionId(self):
		return self.__sectionId

	##To string
	def __str__(self):
		return 'PrefFileItem(id=%s, type=%s, value=%s, sectionId=%s)' % (self.__id, self.__type, self._value, self.__sectionId)

		
##Dummy preference item to fake an preffile item
class PrefFileItemDummy(PrefFileItem):
	##Constructor
	def __init__(self):
		#super constructor
		PrefFileItem.__init__(self, 'dummy', 'dummy', PrefFileItem.TYPE_STRING, '')
		
		#by default no value
		self._value = None

	##Value setter
	def setValue(self, value):
		self._value = value
		
	##Get value
	def getValue(self):
		return self._value
		
##String format preference item
class PrefFileItemString(PrefFileItem):
	##Constructor
	def __init__(self, sectionId, id, type, value):
		PrefFileItem.__init__(self, sectionId, id, type, value)

	##Value setter
	def setValue(self, value):
		#nothing to check
		self._value = str(value)
		
	##Get value
	def getValue(self):
		return str(self._value)

##Integer format preference item
class PrefFileItemInt(PrefFileItem):
	##Constructor
	def __init__(self, sectionId, id, type, value):
		PrefFileItem.__init__(self, sectionId, id, type, value)

	##Value setter
	def setValue(self, value):
		try:
			self._value = int(value)
		except ValueError:
			raise PrefFileItemException('Incorrect TYPE_INT for value "%s"' % str(value))
			
	##Get value
	def getValue(self):
		return int(self._value)

##Float format preference item
class PrefFileItemFloat(PrefFileItem):
	##Constructor
	def __init__(self, sectionId, id, type, value):
		PrefFileItem.__init__(self, sectionId, id, type, value)

	##Value setter
	def setValue(self, value):
		try:
			self._value = float(value)
		except ValueError:
			raise PrefFileItemException('Incorrect TYPE_FLOAT for value "%s"' % str(value))
			
	##Get value
	def getValue(self):
		return float(self._value)

##Date format preference item
class PrefFileItemDate(PrefFileItem):
	##Constructor
	def __init__(self, sectionId, id, type, value):
		#constants
		self.__SEPARATOR = '/'

		#members
		self.__year = 0
		self.__month = 0
		self.__day = 0

		#super constructor
		PrefFileItem.__init__(self, sectionId, id, type, value)

	##Value setter
	def setValue(self, value):
		try:
			#split value
			year,month,day = value.split(self.__SEPARATOR)
			#check date
			date = datetime.date(int(year), int(month), int(day))
			#all seems correct, save value
			self.__year = int(year)
			self.__month = int(month)
			self.__day = int(day)
			self._value = '%d%s%d%s%d' % (self.__year, self.__SEPARATOR, self.__month, self.__SEPARATOR, self.__day)
		except:
			raise PrefFileItemException('Incorrect TYPE_DATE for value "%s"' % str(value))

	##Get year
	def getYear(self):
		return self.__year

    ##Get month
	def getMonth(self):
		return self.__month

    ##Get day
	def getDay(self):
		return self.__day

##Color format preference item
class PrefFileItemColor(PrefFileItem):
	##Constructor
	def __init__(self, sectionId, id, type, value):
		#constants
		self.__SEPARATOR = ':'

		#super constructor
		PrefFileItem.__init__(self, sectionId, id, type, value)

	##Value setter
	def setValue(self, value):
		try:
			#try to convert value in gtk color
			gtk.gdk.color_parse(str(value))
			#it seems to be a correct color
			self._value = value
		except Exception,e :
			raise PrefFileItemException('Incorrect TYPE_COLOR for value "%s"' % str(value))
	
##Boolean format preference item
class PrefFileItemBoolean(PrefFileItem):
	##Constructor
	def __init__(self, sectionId, id, type, value):
		#super constructor
		PrefFileItem.__init__(self, sectionId, id, type, value)

	##Value setter
	def setValue(self, value):
		try:
			if value==True or value=='True' or value=='1':
				self._value = True
			else:
				self._value = False
		except ValueError:
			raise PrefFileItemException('Incorrect TYPE_BOOLEAN for value "%s"' % str(value))
			
	##Get value
	def getValue(self):
		return bool(self._value)
